# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import importlib
import inspect
import pkgutil
from typing import Type, TypeVar

from wurzel.step.typed_step import TypedStep

T = TypeVar("T", bound=object)


def find_sub_classes(parent: T, package: str = __package__) -> dict[str, T]:
    """searches for all DVC step definitions and returns them based on their name."""

    def is_non_abs_child(member: object) -> bool:
        return (
            True
            and inspect.isclass(member)
            and issubclass(member, parent)
            and not inspect.isabstract(member)
            and not bool(getattr(member, "__abstractmethods__", False))
        )

    result = {}
    visited = set([f"{__package__}.main", f"{__package__}.utils"])  # noqa: C405
    module_iterator = pkgutil.iter_modules(importlib.import_module(package).__path__)
    for _, module_name, is_package in module_iterator:
        full_module_name = f"{package}.{module_name}"
        if full_module_name in visited:
            continue
        visited.add(full_module_name)
        # Recurse through any sub-packages
        if is_package:
            classes_in_subpackage = find_sub_classes(parent, package=full_module_name)
            result.update(classes_in_subpackage)
        # Load the module for inspection
        module = importlib.import_module(full_module_name)

        # Iterate through all the objects in the module and
        # using the lambda, filter for class objects and only objects that exist within the module
        for _name, obj in inspect.getmembers(module, is_non_abs_child):
            result[obj.__name__] = obj
    return result


def find_typed_steps_in_package(package: str) -> dict[str, Type[TypedStep]]:
    """Recursively find all subclasses of TypedStep"""
    return find_sub_classes(TypedStep, package)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import abc
from collections.abc import Iterable
from logging import getLogger
from pathlib import Path
from types import NoneType
from typing import (
    Generic,
    Optional,
    Self,
    Type,
    TypeAlias,
    TypeVar,
    Union,
    get_args,
)

import pandera.typing as patyp
from typing_inspect import get_origin

from wurzel.datacontract import PanderaDataFrameModel, PydanticModel
from wurzel.exceptions import ContractFailedException, StaticTypeError
from wurzel.path import PathToFolderWithBaseModels
from wurzel.step.settings import Settings
from wurzel.step.step import Step

# pylint: disable-next=invalid-name
MODEL_TYPE: TypeAlias = Type[Union[PydanticModel, PanderaDataFrameModel]]
#  ^Should be a Intersection between DataModel & {BaseModel, DataFrameModel}
log = getLogger(__name__)
INCONTRACT = TypeVar("INCONTRACT")
OUTCONTRACT = TypeVar("OUTCONTRACT")


SETTS = TypeVar("SETTS", bound=Settings)


class TypedStep(Step, Generic[SETTS, INCONTRACT, OUTCONTRACT]):
    """Wurzel Pipeline Step.
    Makes use of **strong** typing. i.e. the types are not only hints.
    On `__init__` it will check annotations to make sure everything is correct.
    The only method that needs implementation is the run method.
    # Description:
    A (Knowledge / Wurzel) pipeline consists of many steps chained together.
    First data is gathered, then (>>) transformed etc.

    Each step (TypedStep) is defined by a triplet of Optional[Settings], Optional[input_d_type], output_d_type.
    Only if the output of a given step is compatible with the next, chaining is allowed.
    ## Settings
    The first part of the triplet. Is either None or a Subtype of wurzel.Settings (internally StepSettings).
    These settings are a special form of pydantic settings meaning they will try to load
    their values from the environment with the prefix `UPPERCASE_STEP_NAME`.
    For more information take a look at wurzel.step.settings
    ## Contract / Input-Output Classes
    The Inputs and outputs of TypedSteps are restricted.
    However they need to inherit from  wurzel.datacontract.DataModel
    There are already two subclasses in wurzel.datacontract implemented.
    Those extend both pydantic and pandera for typesafe objects and dataframes.
    Containers such as lists for both types are supported
    ### Input
    If the Input of a step is set to be None the step is a leaf and will always run.
    The `inputs` arg in the run method of the step can be ignored.
    ### Ouput
    The Step output can't be None.
    ### Input
    ## Example:
    ```python
    from wurzel.step import (
        PanderaDataFrameModel,
        TypedStep,
        PydanticModel,
        NoSettings
        )
    from pandera.typing import Series

    # Data Model definitions
    class MyData(PydanticModel):
        key: str = "Value"
    class MyTable(PanderaDataFrameModel):
        col1: Series[int]

    # Step definition
    class MyStep(TypedStep[NoSettings, None, list[MyData]]):
        def run(self, inpt: None) -> list[MyData]:
            return ...
    class MyOtherStep(TypedStep[None, list[MyData], MyTable]):
        def run(self, inpt: list[MyData]) -> MyTable:
            return ...
    ```
    """

    _internal_input_class: Type[PathToFolderWithBaseModels]
    _internal_output_class: Type[PathToFolderWithBaseModels]
    input_model_type: Union[MODEL_TYPE, list[MODEL_TYPE], None]
    output_model_type: Union[MODEL_TYPE, list[MODEL_TYPE], None]
    settings_class: Type[SETTS]
    output_model_class: MODEL_TYPE
    input_model_class: MODEL_TYPE
    _supported_containers: Iterable[Type[Iterable]] = (list, set, patyp.DataFrame)
    settings: SETTS

    def output_path(self, folder: Path) -> Path:
        """Used in generate dvc yml. Creates a path"""
        pth = folder / self.__class__.__name__
        return pth.with_suffix(self.output_model_class.kt_file_extension())

    @classmethod  #
    def __unpack_list_containers(
        cls,
        list_or_type: Union[list, Type, None],
        containers: Optional[list[Type[Iterable]]] = None,
    ) -> tuple[Optional[Iterable[Type[Iterable]]], Type]:
        """Unpacks the containers around a given nested Type

        Args:
            list_or_type (Union[list, Type, None]): Type or container to try to unpack
            containers (Optional[list[Type[Iterable]]], optional): used in recursive calls. Defaults to None.

        Raises:
            StaticTypeError: Anyting goes wrong

        Returns:
            tuple[Optional[Iterable[Type[Iterable]]], Type]: (container types, type)
        """
        if containers is None:
            containers = []
        origin_t: Union[Type[Iterable], Type, None] = get_origin(list_or_type)
        if origin_t is None:
            # list_or_type is now a type/class
            if list_or_type is None:
                list_or_type = NoneType
            return containers, list_or_type  # type: ignore[return-value] # list_or_type is now a type.
        if origin_t in cls._supported_containers:
            containers.insert(0, origin_t)
        else:
            raise StaticTypeError(f"{origin_t} is not a supported container")
        if len(get_args(list_or_type)) == 1:
            # Unpack list_type
            return cls.__unpack_list_containers(get_args(list_or_type)[0], containers)
        raise StaticTypeError(f"Cant handle type with {len(get_args(origin_t))} args")

    @classmethod  # pylint: disable-next=unused-private-member # used in __new__
    def __static_type_check_self(cls):
        """Confirms the type annotations of child classes TypedStep"""
        type_annotations = [
            get_args(parent)
            for parent in getattr(cls, "__orig_bases__", [])
            if get_args(parent)
        ][0]
        if type_annotations == ():
            raise StaticTypeError(
                f"No type-annotation provided when creating subclass of {cls.__name__}"
                + f"Use: MyStep({cls.__name__}[INPUT_T, OUTPUT_T])"
            )
        cls.settings_class, cls.input_model_type, cls.output_model_type = (
            type_annotations
        )
        if not issubclass(cls.settings_class, (Settings, NoneType)):
            raise StaticTypeError(
                "Settings provided in TypedStep[<>, ...]"
                + " is not a subclass of settings_class"
            )
        _ = cls.__unpack_list_containers(cls.input_model_type)
        out_t = cls.__unpack_list_containers(cls.output_model_type)

        def has_no_annotation(c: list, t: type):
            return c == [] and t == NoneType

        if has_no_annotation(*out_t):
            raise StaticTypeError(
                f"Type-annotation for output of {cls.__name__}[..., None] can't be None"
            )

    @classmethod  # pylint: disable-next=unused-private-member # used in __new__
    def __static_type_check_run(cls):
        """Confirms the type annotations of child classes run method.
        Currently allows missing type annotations with a warning
        """
        # Since type annotation of class does not want list container
        # but run adds it, we add it manually
        expected_run_input = cls.input_model_type
        expected_signature_str = (
            f"run(inputs: {expected_run_input}) -> {cls.output_model_type}:"
        )
        try:
            if not cls.run.__annotations__:
                log.warning(
                    f"The step {cls.__name__} has no types. This is not recommended!"
                )
                log.info(
                    f"Method signature should be {cls.__name__}.{expected_signature_str}:"
                )
                # raise StaticTypeError(
                #    "incorrect function signature for run method:  no typing"
                # )
                annotations = {
                    "return": cls.output_model_type,
                    "inputs": expected_run_input,
                }
            else:
                annotations = cls.run.__annotations__.copy()
            run_retur = annotations.pop("return")
            if len(annotations) != 1:
                raise StaticTypeError(
                    "incorrect funtion signature (inputs) for run method: "
                    + f"Expected one input, got {annotations}"
                )
            _, run_input = annotations.popitem()
            run_input_cons, run_input_orig = cls.__unpack_list_containers(run_input)
            run_retur_cons, run_retur_orig = cls.__unpack_list_containers(run_retur)
            # construct type using only list instead of List
            return_annotation = run_retur_orig
            for container in run_retur_cons:
                return_annotation = container[return_annotation]
            input_annotation = run_input_orig
            for container in run_input_cons:
                input_annotation = container[input_annotation]
            # Check if inputs was in list
            if input_annotation != expected_run_input:
                raise StaticTypeError(
                    "Incorrect function signature (inputs) for run method:\n"
                    + f"\tis       run({run_input}) -> ...\n"
                    + f"\texpected {expected_signature_str}"
                )
            if return_annotation != cls.output_model_type:
                raise StaticTypeError(
                    "Incorrect function signature (return) for run method:\n"
                    + f"\tis       run(...) -> {run_retur.__name__}\n"
                    + f"\texpected {expected_signature_str}"
                )
        except IndexError as i:
            raise ContractFailedException(
                "Could not get_args of either run inputs or return"
            ) from i
        except KeyError as k:
            raise ContractFailedException("Could not get annotations from run") from k

    def finalize(self) -> None:
        """Called after execution in Executor finished

        Can be used for cleanup etc.

        One Example would be the retirement of collections in a db step.
        """

    def __new__(cls) -> Self:
        instance = super(TypedStep, cls).__new__(cls)
        super().__init__(instance)
        # Get Input and output type from annotations
        # Get Type annotations, fallback to None if they don't exist
        # pylint: disable-next=no-member
        instance.__static_type_check_self()

        instance.input_model_class = (
            get_args(instance.input_model_type) or [instance.input_model_type]
        )[-1]
        instance.output_model_class = (
            get_args(instance.output_model_type) or [instance.output_model_type]
        )[-1]

        instance.__static_type_check_run()

        # Sadly we cant use type() or types.new_class for this.
        class InCls(PathToFolderWithBaseModels[instance.input_model_type]):
            """Used internally"""

        class OutCls(PathToFolderWithBaseModels[instance.output_model_type]):
            """Used internally"""

        instance._internal_input_class = InCls
        instance._internal_output_class = OutCls
        return instance

    # super was called in __new__
    # pylint: disable-next=super-init-not-called
    def __init__(self) -> None:
        self.settings = self.settings_class()

    def add_required_step(self, step: "TypedStep"):
        """Add step to execution graph

        Args:
            step (TypedStep) Step to add

        Raises:
            TypeError: On incompatible types
        """
        if self.input_model_type != step.output_model_type:
            raise TypeError(
                f"Cannot chain {self} to {step} ({step.output_model_type} -> {self.input_model_type})"
            )
        super().add_required_step(step)

    def _traverse(self, set_of: set["TypedStep"]):
        set_of.add(self)
        for step in self.required_steps:
            TypedStep._traverse(step, set_of)
        return set_of

    def traverse(self) -> set["TypedStep"]:
        """Retrieve a set of all required steps
        including self"""
        return self._traverse(set())

    # pylint: disable=method-hidden
    @abc.abstractmethod
    def run(self, inpt: INCONTRACT) -> OUTCONTRACT:
        """abstract function which is called with the parsed data in the shape of INCONTRACT.
        It's implementation should return the OUTCONTRACT
        ### It will be called mutliple times.

        """
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""CLI program"""

import importlib.metadata
from pathlib import Path
from typing import Type

from wurzel.step import TypedStep
from wurzel.step_executor import BaseStepExecutor

__all__ = ["generate_cli_call"]
try:
    __version__ = importlib.metadata.version("wurzel")
# pylint: disable-next=bare-except
except:  # noqa: E722
    __version__ = "dev"
__version_info__ = __version__.split(".")


def generate_cli_call(
    step_cls: Type[TypedStep],
    inputs: list[Path],
    output: Path,
    executor: Type[BaseStepExecutor] = None,
    encapsulate_env: bool = True,
) -> str:
    """Generate the cli call to execute a given step with its
    inputs and output

    Args:
        step_cls (Type[TypedStep]): Step to execute
        inputs (list[Path]): list of Directories
        output (Path): Output Directory

    Returns:
        str: cmd
    """
    if inputs:
        inputs = "-i " + " -i ".join(i.as_posix() for i in inputs)
    else:
        inputs = ""
    return " ".join(
        [
            "python3 -m wurzel run",
            f"{step_cls.__module__}:{step_cls.__qualname__}",
            "-o",
            output.as_posix(),
            "" if executor is None else f"-e {executor.__qualname__}",
            inputs,
            "--encapsulate-env" if encapsulate_env else "--no-encapsulate-env",
        ]
    )



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""CLI Entry"""

import importlib
import inspect
import logging
import logging.config
import os
import pkgutil
import sys
from datetime import datetime
from pathlib import Path
from typing import Annotated

import typer
import typer.core

from wurzel.adapters.dvc_adapter import DvcBackend
from wurzel.cli.cmd_generate import main as cmd_generate
from wurzel.cli.cmd_inspect import main as cmd_inspect
from wurzel.cli.cmd_run import main as cmd_run
from wurzel.step import TypedStep
from wurzel.step_executor import BaseStepExecutor, PrometheusStepExecutor
from wurzel.utils.logging import get_logging_dict_config
from wurzel.utils.meta_settings import WZ
from wurzel.utils.meta_steps import find_sub_classes

app = typer.Typer(
    no_args_is_help=True,
)
log = logging.getLogger(__name__)


def executer_callback(_ctx: typer.Context, _param: typer.CallbackParam, value: str):
    """Convert a cli-str to a Type[BaseStepExecutor]

    Args:
        _ctx (typer.Context)
        _param (typer.CallbackParam):
        value (str): user typed string

    Raises:
        typer.BadParameter: If user typed string does not correlate with a Executor

    Returns:
        Type[BaseStepExecutor]: {BaseStepExecutor, PrometheusStepExecutor}
    """
    if "BASESTEPEXECUTOR".startswith(value.upper()):
        return BaseStepExecutor
    if "PROMETHEUSSTEPEXECUTOR".startswith(value.upper()):
        return PrometheusStepExecutor
    raise typer.BadParameter(f"{value} is not a recognized executor")


def step_callback(
    _ctx: typer.Context, _param: typer.CallbackParam, import_path: str
) -> TypedStep:
    """Converts a cli-str to a TypedStep

    Args:
        _ctx (typer.Context):
        _param (typer.CallbackParam):
        path (str): user-typed string

    Raises:
        typer.BadParameter: import not possible

    Returns:
        Type[TypedStep]: <<step>>
    """
    try:
        if ":" in import_path:
            mod, kls = import_path.rsplit(":", 1)
        else:
            mod, kls = import_path.rsplit(".", 1)
        module = importlib.import_module(mod)
        step = getattr(module, kls)
        assert (inspect.isclass(step) and issubclass(step, TypedStep)) or isinstance(
            step, TypedStep
        )
    except ValueError as ve:
        raise typer.BadParameter(
            "Path is not in correct format, should be module.submodule.Step"
        ) from ve
    except ModuleNotFoundError as me:
        raise typer.BadParameter(f"Module '{mod}' could not be imported") from me
    except AttributeError as ae:
        raise typer.BadParameter(f"Class '{kls}' not in module {module}") from ae
    except AssertionError as ae:
        raise typer.BadParameter(f"Class '{kls}' not a TypedStep") from ae
    return step


def complete_step_import(incomplete: str):
    """AutoComplete for steps"""

    packages = [
        p
        for p in pkgutil.iter_modules()
        if p.ispkg and p.name.startswith(incomplete if incomplete else "wurzel")
    ]
    hints = []
    for pkg in packages:
        hints.extend(
            ".".join([cls.__module__, cls.__qualname__])
            for cls in find_sub_classes(TypedStep, pkg.name).values()
            if str(".".join([cls.__module__, cls.__qualname__])).startswith(incomplete)
        )

    logging.info("found possible steps:", extra={"packages": packages, "hints": hints})
    return hints


@app.command(no_args_is_help=True, help="Run a step")
# pylint: disable-next=dangerous-default-value
def run(
    step: Annotated[
        str,
        typer.Argument(
            allow_dash=False,
            help="module path to step",
            autocompletion=complete_step_import,
            callback=step_callback,
        ),
    ],
    output_path: Annotated[
        Path,
        typer.Option("-o", "--output", file_okay=False, help="Folder with outputs"),
    ] = Path(f"<step-name>-{datetime.now().isoformat(timespec='milliseconds')}"),
    input_folders: Annotated[
        list[Path],
        typer.Option(
            "-i",
            "--inputs",
            help="input folders",
            file_okay=False,
            exists=True,
        ),
    ] = [],
    executor: Annotated[
        str,
        typer.Option(
            # "",
            "-e",
            "--executor",
            help="executor to use",
            callback=executer_callback,
            autocompletion=lambda: ["BaseStepExecutor", "PrometheusStepExecutor"],
        ),
    ] = "BaseStepExecutor",
    encapsulate_env: Annotated[bool, typer.Option()] = True,
):
    """run"""
    output_path = Path(output_path.as_posix().replace("<step-name>", step.__name__))
    log.debug(
        "executing run",
        extra={
            "parsed_args": {
                "step": step,
                "output_path": output_path,
                "input_folders": input_folders,
                "executor": executor,
                "encapsulate_env": encapsulate_env,
            }
        },
    )
    return cmd_run(step, output_path, input_folders, executor, encapsulate_env)


@app.command("inspect", no_args_is_help=True, help="Display information about a step")
def inspekt(
    step: Annotated[
        str,
        typer.Argument(
            allow_dash=False,
            help="module path to step",
            autocompletion=complete_step_import,
            callback=step_callback,
        ),
    ],
    gen_env: Annotated[bool, typer.Option()] = False,
):
    """inspect"""

    return cmd_inspect(step, gen_env)


def backend_callback(_ctx: typer.Context, _param: typer.CallbackParam, _backend: str):
    """validates input and returns fitting backend. Currently always DVCBackend"""
    logging.warning("only DVCBackend is supported currently")
    return DvcBackend


def pipeline_callback(
    _ctx: typer.Context, _param: typer.CallbackParam, import_path: str
) -> TypedStep:
    """Based on step_callback transform them to WZ pipeline elements"""
    step = step_callback(_ctx, _param, import_path)
    if not hasattr(step, "required_steps"):
        step = WZ(step)
    return step


@app.command(no_args_is_help=True, help="generate a pipeline")
# pylint: disable-next=dangerous-default-value
def generate(
    pipeline: Annotated[
        str,
        typer.Argument(
            allow_dash=False,
            help="module path to step or pipeline(which is a chained step)",
            autocompletion=complete_step_import,
            callback=pipeline_callback,
        ),
    ],
    data_dir: Annotated[
        Path,
        typer.Option(
            "-d", "--data-dir", file_okay=False, help="Target folder for pipeline"
        ),
    ] = Path("./data"),
    backend: Annotated[
        str,
        typer.Option(
            "-b",
            "--backend",
            callback=backend_callback,
            help="backend to use",
        ),
    ] = DvcBackend,
):
    """run"""
    log.debug(
        "generate pipeline",
        extra={
            "parsed_args": {
                "pipeline": pipeline,
                "data_dir": data_dir,
                "backend": backend,
            }
        },
    )
    return print(
        cmd_generate(
            pipeline,
            data_dir,
            backend=backend,
        )
    )


def update_log_level(log_level: str):
    """Fix for typer logs"""
    log_config = get_logging_dict_config(log_level)
    log_config["formatters"]["default"] = {
        "()": "wurzel.cli.logger.WithExtraFormatter",
        "reduced": ["INFO"],
    }
    log_config["handlers"]["default"] = {
        "()": "rich.logging.RichHandler",
        "formatter": "default",
    }
    logging.config.dictConfig(log_config)


@app.callback()
def main_args(
    verbose: Annotated[bool, typer.Option("--verbose")] = False,
    log_level: Annotated[
        str,
        typer.Option(
            "--log-level",
            autocompletion=lambda: ["CRITICAL", "FATAL", "ERROR", "WARN", "INFO"],
        ),
    ] = "INFO",
):
    """global settings, main"""
    if not os.isatty(1):
        typer.core.rich = None
        logging.config.dictConfig(get_logging_dict_config(log_level))
        app.pretty_exceptions_enable = False
        app.pretty_exceptions_show_locals = False
    else:
        update_log_level(log_level)
        app.pretty_exceptions_enable = True
        app.pretty_exceptions_show_locals = True
        app.pretty_exceptions_short = not verbose


def main():
    """main"""
    sys.path.append(os.getcwd())  # needed fo find the files relative to cwd
    app()



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from importlib.util import find_spec as _find_spec
from logging import getLogger
from typing import Any

from . import semantic_splitter
from .meta_settings import WZ, create_model
from .meta_steps import find_typed_steps_in_package
from .to_markdown.html2md import to_markdown

log = getLogger(__name__)

_opt_deps = {k: bool(_find_spec(k)) for k in ["tlsh", "pymilvus", "qdrant_client"]}

HAS_TLSH = _opt_deps["tlsh"]
HAS_MILVUS = _opt_deps["pymilvus"]
HAS_QDRANT = _opt_deps["qdrant_client"]
log.info("Optional deps in env", extra={**_opt_deps})
__all__ = [
    "semantic_splitter",
    "WZ",
    "create_model",
    "try_get_length",
    "find_typed_steps_in_package",
    "to_markdown",
    "HAS_MILVUS",
    "HAS_TLSH",
]


def try_get_length(x: Any) -> int:
    """Tries to get length, return 1 if fails

    Args:
        x (Any): get length of

    Returns:
        int: length or 1
    """
    try:
        return len(x)
    # pylint: disable=bare-except
    except:  # noqa: E722
        return 1



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""containing the DVCStep sending embedding data into Qdrant"""

# pylint: disable=duplicate-code
import itertools
from logging import getLogger

from pandera.typing import DataFrame
from qdrant_client import models

from wurzel.exceptions import StepFailed
from wurzel.step import TypedStep
from wurzel.steps.embedding.data import EmbeddingMultiVectorResult
from wurzel.steps.qdrant.step import QdrantConnectorStep

from .data import QdranttMultiVectorResult
from .settings import QdrantSettings

log = getLogger(__name__)


def _batch(iterable, size):
    it = iter(iterable)
    while item := list(itertools.islice(it, size)):
        yield item


class QdrantConnectorMultiVectorStep(
    QdrantConnectorStep,
    TypedStep[
        QdrantSettings,
        DataFrame[EmbeddingMultiVectorResult],
        DataFrame[QdranttMultiVectorResult],
    ],
):
    """
    Qdrant connector step. It consumes embedding csv files, creates a new schema and inserts the embeddings
    """

    def _create_collection(self, size: int):
        self.client.create_collection(
            collection_name=self.collection_name,
            vectors_config=models.VectorParams(
                size=size,
                distance=self.settings.DISTANCE,
                multivector_config=models.MultiVectorConfig(
                    comparator=models.MultiVectorComparator.MAX_SIM
                ),
            ),
            replication_factor=self.settings.REPLICATION_FACTOR,
        )

    def run(
        self, inpt: DataFrame[EmbeddingMultiVectorResult]
    ) -> DataFrame[QdranttMultiVectorResult]:
        if not self.client.collection_exists(self.collection_name):
            self._create_collection(len(inpt["vectors"].loc[0][0]))
        df_result = self._insert_embeddings(inpt)
        return df_result

    def _insert_embeddings(
        self, data: DataFrame[EmbeddingMultiVectorResult]
    ) -> DataFrame[QdranttMultiVectorResult]:
        log.debug(f"Creating Qdrant collection {self.collection_name}")

        log.info(
            "Inserting embeddings",
            extra={"count": len(data), "collection": self.collection_name},
        )
        points = [
            models.PointStruct(
                id=next(self.id_iter),  # type: ignore[arg-type] # No MultiIndex, so always int.
                vector=row["vectors"],
                payload={
                    "url": row["url"],
                    "text": row["text"],
                    **self.get_available_hashes(row["text"]),
                    "splits": row["splits"],
                    "keywords": row["keywords"],
                },
            )
            for i, row in data.iterrows()
        ]
        for point_chunk in _batch(points, self.settings.BATCH_SIZE):
            operation_info = self.client.upsert(
                collection_name=self.collection_name, wait=True, points=point_chunk
            )
            if operation_info.status != "completed":
                raise StepFailed(
                    f"Failed to insert df chunk into collection '{self.collection_name}' {operation_info}"
                )
            log.info(
                "Successfully inserted vector_chunk",
                extra={"collection": self.collection_name, "count": len(point_chunk)},
            )
        self._create_indices()
        self._update_alias()
        data = [
            {
                **entry.payload,
                "vectors": entry.vector,
                "collection": self.collection_name,
                "id": entry.id,
            }
            for entry in points
        ]
        result = DataFrame[QdranttMultiVectorResult](data)
        return result



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .common import MarkdownDataContract
from .datacontract import DataModel, PanderaDataFrameModel, PydanticModel

__all__ = [
    "PanderaDataFrameModel",
    "PydanticModel",
    "MarkdownDataContract",
    "DataModel",
]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from ..datacontract import MarkdownDataContract, PanderaDataFrameModel, PydanticModel
from ..path import PathToFolderWithBaseModels
from ..step_executor import BaseStepExecutor, PrometheusStepExecutor
from .history import step_history  # noqa: F401
from .settings import NoSettings, Settings
from .step import Step
from .typed_step import TypedStep

# pylint: disable-next=invalid-name

__all__ = [
    "TypedStep",
    "Step",
    "step_history",
    "PanderaDataFrameModel",
    "PydanticModel",
    "MarkdownDataContract",
    "PathToFolderWithBaseModels",
    "NoSettings",
    "Settings",
    "BaseStepExecutor",
    "PrometheusStepExecutor",
]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import inspect
import json
from collections.abc import Iterable
from contextvars import ContextVar
from types import NoneType
from typing import Optional, Union

from .typed_step import TypedStep

step_history: ContextVar[Optional["History"]] = ContextVar("step_history", default=None)
step_history.set(None)


class History:
    """internal history object"""

    __SEP_STR = "-"
    _history: list[str]

    def __init__(
        self, *args: Union[TypedStep, str, list[str]], initial: list[str] = None
    ) -> NoneType:
        if initial is None:
            initial = []
        self._history = initial
        self += args

    def __add(self, s: str):
        self._history.append(s[:-4] if s.endswith("Step") else s)

    def __iadd__(self, other: Union[TypedStep, str, list[str]]):
        if isinstance(other, str):
            self.__add(other)
        elif isinstance(other, Iterable):
            for i in other:
                self.__iadd__(i)
        elif inspect.isclass(other):
            self.__add(other.__name__)
        elif isinstance(other, object):
            self.__add(other.__class__.__name__)
        else:
            self.__add(str(other))
        return self

    def copy(self) -> "History":
        """Returns a copy of self"""
        return History(initial=self.get())

    def get(self) -> list[str]:
        """get History

        Returns:
            list[str]: history (copy)
        """
        return self._history.copy()

    def __add__(self, other):
        if isinstance(other, History):
            return History(initial=[*self._history, *other._history])
        cpy = self.copy()
        cpy += other
        return cpy

    def __str__(self) -> str:
        return History.__SEP_STR.join(self._history)

    def __repr__(self) -> str:
        return f"History({self._history})"

    def __getitem__(self, key: Union[str, int, slice]):
        if isinstance(key, str):
            raise TypeError
        return History(self._history[key])

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, History):
            return False
        return self._history == value._history

    def to_json(self):
        """converts to json string

        Returns:
            str: json list of str
        """
        return json.dumps(self._history)

    @classmethod
    def from_json(cls, s: str) -> "History":
        """converts from json string

        Args:
            s (str): json string list of str

        Returns:
            History: new Instance
        """
        return History(initial=json.loads(s))

    @classmethod
    def from_str(cls, s: str) -> "History":
        """converts from str(History)

        Args:
            s (str)

        Returns:
            History: new Instance
        """
        return History(initial=json.loads(s))



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""Semantic Markdown Splitter"""

import re
from logging import getLogger
from typing import TYPE_CHECKING, Optional, Tuple, Type, TypedDict

import mdformat
import tiktoken
from mistletoe import Document as MisDocument
from mistletoe import block_token, markdown_renderer, span_token
from mistletoe.token import Token

from wurzel.datacontract import MarkdownDataContract
from wurzel.exceptions import MarkdownException

if TYPE_CHECKING:
    import spacy
LEVEL_MAPPING = {
    block_token.Heading: 0,  # actually 1-6
    block_token.List: 7,
    block_token.CodeFence: 8,
    block_token.Table: 9,
    block_token.Paragraph: 10,
    block_token.ListItem: 11,
    block_token.TableRow: 12,
    span_token.LineBreak: 13,
    block_token.ThematicBreak: 14,
}
LEVEL_UNDEFINED = 15
OPENAI_ENCODING = tiktoken.encoding_for_model("gpt-3.5-turbo")

log = getLogger(__name__)


class MetaDataDict(TypedDict):
    """Dict definition of metadata"""

    keywords: str
    url: str


class DocumentNode(TypedDict):
    """Dict definition of internal tree structure for splitting"""

    highest_level: int
    token_len: int
    text: str
    children: list["DocumentNode"]
    metadata: MetaDataDict


def _get_token_len(text: str) -> int:
    """Get OpenAI Token length

    Args:
        text (str): Test to encode

    Returns:
        int: count of tokens
    """
    return len(OPENAI_ENCODING.encode(text))


def _is_all_children_same_level(children: list[DocumentNode]) -> bool:
    nbr_different_level = len({c["highest_level"] for c in children})
    return nbr_different_level == 1


def _get_children_sorted_by_level(
    children: list[DocumentNode],
) -> list[tuple[int, int]]:
    return sorted(
        [(i, c["highest_level"]) for i, c in enumerate(children)],
        key=lambda x: x[1],
        reverse=False,
    )


def _get_highest_index_of_children(children: list[DocumentNode]) -> int:
    sorted_by_level = _get_children_sorted_by_level(children)
    return sorted_by_level[0][0]


def _get_highest_level_of_children(children: list[DocumentNode]) -> int:
    sorted_by_level = _get_children_sorted_by_level(children)
    return sorted_by_level[0][1]


def _get_heading_text(token: block_token.Heading):
    """Get heading text from Mistletoe heading block token"""
    if token.content:
        return token.content
    if "children" in vars(token) and len(token.children) > 0:
        return " ".join([c.content for c in token.children if hasattr(c, "content")])
    return ""


def _is_standalone_a_heading(text):
    childs = MisDocument(text).children
    if len(childs) != 1:
        return False
    return isinstance(childs[0], block_token.Heading)


def _cut_to_tokenlen(text: str, token_len: int) -> str:
    """Cut Text to token length using OpenAI"""

    tokens_old = OPENAI_ENCODING.encode(text)
    if len(tokens_old) > token_len:
        tokens = tokens_old[:token_len]
        return OPENAI_ENCODING.decode(tokens)
    return OPENAI_ENCODING.decode(tokens_old)


def _format_markdown_docs(
    docs: list[MarkdownDataContract],
) -> list[MarkdownDataContract]:
    """Formats the Markdown Document by the standards"""
    return [
        MarkdownDataContract(
            md=mdformat.text(doc.md).strip(),
            url=doc.url,
            keywords=doc.keywords,
        )
        for doc in docs
    ]


class WurzelMarkdownRenderer(markdown_renderer.MarkdownRenderer):
    """Fix For markdown_renderer.MarkdownRenderer"""

    # pylint: disable=unused-argument, arguments-differ
    def render_table_cell(
        self, token: block_token.TableCell, max_line_length: int
    ) -> str:
        return self.render_inner(token)

    def render_table_row(
        self, token: block_token.TableRow, max_line_length: int
    ) -> str:
        return self.render_inner(token)


class SemanticSplitter:
    """Splitter implementation"""

    nlp: "spacy.language.Language"
    token_limit: int
    token_limit_buffer: int
    token_limit_min: int

    def __init__(
        self,
        token_limit: int = 256,
        token_limit_buffer: int = 32,
        token_limit_min: int = 64,
        spacy_model: str = "de_core_news_sm",
    ) -> None:
        import spacy  # pylint: disable=import-outside-toplevel

        self.nlp = spacy.load(spacy_model)
        self.token_limit = token_limit
        self.token_limit_buffer = token_limit_buffer
        self.token_limit_min = token_limit_min

    def _is_short(self, text: str) -> bool:
        return _get_token_len(text) <= self.token_limit - self.token_limit_buffer

    def _is_table(self, doc: DocumentNode) -> bool:
        return doc["highest_level"] == LEVEL_MAPPING[block_token.Table]

    def _is_within_targetlen_w_buffer(self, text: str) -> bool:
        length = _get_token_len(text)
        return (
            self.token_limit + self.token_limit_buffer
            >= length
            >= self.token_limit - self.token_limit_buffer
        )

    def _render_doc(self, doc: MisDocument) -> str:
        """Render Mistletoe Markdown Document"""
        try:
            with WurzelMarkdownRenderer() as renderer:
                return renderer.render(doc)  # type: ignore[no-any-return]
        except Exception as e:
            raise MarkdownException(e) from e

    def _get_custom_level(self, block: block_token.BlockToken) -> int:
        """Get the hierarchical level for a mistletoe node"""

        if isinstance(block, block_token.Heading):
            return int(block.level)
        return LEVEL_MAPPING.get(type(block), LEVEL_UNDEFINED)

    def _merge_children(self, children: list[DocumentNode]) -> MisDocument:
        """Create a document out of a list of children"""
        new_doc = MisDocument([])
        # If all children a span tokens add them to a paragraph
        # because problems otherwise
        if all(isinstance(c, span_token.SpanToken) for c in children):
            para = block_token.Paragraph([])
            para.children = children
            new_doc.children += [para]
        else:
            new_doc.children += children
        return new_doc

    def _find_highest_level(
        self, children: list[DocumentNode], min_level: int = 0
    ) -> Tuple[int, Optional[Token], Optional[DocumentNode]]:
        """
        Among a list of children nodes find the one with the highest level.
        Return a tuple of that level, level node type and that child
        """

        def is_any_children(
            children: list[DocumentNode], block_type: Type[block_token.BlockToken]
        ):
            """Check if any Mistletoe Node (child) is of specific type"""
            for child in children:
                if isinstance(child, block_type):
                    return True
            return False

        highest_level: int = LEVEL_UNDEFINED
        highest_type: Optional[Type[Token]] = None
        highest_element: Optional[DocumentNode] = None

        if children is None:
            return (LEVEL_UNDEFINED, None, None)
        # Checked by higher level function
        for child in children:
            if isinstance(child, block_token.Heading) and child.level > min_level:
                if child.level < highest_level:
                    highest_level = int(child.level)
                    highest_type = block_token.Heading
                    highest_element = child

        # block_token.ThematicBreak
        # If level is not set by the Heading
        for block_type in [
            block_token.List,
            block_token.CodeFence,
            block_token.Table,
            block_token.Paragraph,
            block_token.ListItem,
            block_token.TableRow,
            span_token.LineBreak,  # Add table
            block_token.ThematicBreak,
        ]:
            if (
                (highest_level == LEVEL_UNDEFINED)
                and is_any_children(children, block_type)
                and (LEVEL_MAPPING[block_type] > min_level)
            ):
                highest_level = LEVEL_MAPPING[block_type]
                highest_type = block_type
                highest_element = [
                    child for child in children if isinstance(child, block_type)
                ][0]
                break

        return highest_level, highest_type, highest_element

    def _split_children(
        self, children: list[MisDocument], min_level: int = 0
    ) -> list[MisDocument]:
        """
        Split a list of children in the most semantic way
        and return a list of Documents with them merged
        """
        if len(children) == 1:
            if "children" in vars(children[0]) or "_children" in vars(children[0]):
                return self._split_children(children[0].children)
            return [children[0]]
        highest_level, highest_type, _ = self._find_highest_level(children, min_level)

        # No higher splitter found
        if highest_level == LEVEL_UNDEFINED:
            return children
        assert highest_type
        # Find point to split the list of children
        split_points: list[int] = []
        for i, child in enumerate(children):
            if isinstance(child, highest_type):
                if highest_level == self._get_custom_level(child):
                    split_points.append(i)

        # There must be at least one split if highest_level < LEVEL_UNDEFINED
        # Highest splitter is first element
        if (split_points[0] == 0) and len(split_points) == 1:
            # Remove leading line breaks
            if isinstance(children[0], span_token.LineBreak):
                children = children[1:]
            # run again with lower level requirement
            return self._split_children(children, highest_level)

        # Combine Children to Documents
        prev_i = 0
        return_docs = []
        for i in split_points:
            if children[prev_i:i]:
                new_doc = self._merge_children(children[prev_i:i])
                return_docs.append(new_doc)
            prev_i = i
        new_doc = self._merge_children(children[split_points[-1] :])
        return_docs.append(new_doc)
        return return_docs

    def text_sentences(self, text):
        """Split a text into sentences using a NLP model.

        This does not use a Regex based approach on purpose as they break with
        punctuation very easily see: https://stackoverflow.com/a/61254146"""
        return [sentence_span.text for sentence_span in self.nlp(text).sents]

    def _markdown_hierarchy_parser(
        self, text: str, metadata: MetaDataDict, max_depth: int = 30
    ) -> DocumentNode:
        """
        Splits a Markdown string into a semantic Markdown based hierarchy

        Given a Markdown string it hierarchically splits that text using
        the semantic information from the Markdown document until
        all final leaf nodes are below the global token limit.

        max_depth decreases with every recursive call of the function
        controlling the maximum depth of the hierarchy.
        """
        md_doc = MisDocument(text)
        highest_level, _, _ = self._find_highest_level(md_doc.children)
        token_len = _get_token_len(text)

        # Reached max recursion depth
        if max_depth == 0:
            log.warning("maximal markdown recursion reached")
            return DocumentNode(
                highest_level=highest_level,
                token_len=token_len,
                text=text,
                metadata=metadata,
                children=[],
            )

        # Do further hierarchy parsing
        splits: MisDocument = self._split_children(md_doc.children)

        def has_node_a_known_level(x):
            return any(
                isinstance(x.children[0] if isinstance(x, MisDocument) else x, cl)
                for cl in LEVEL_MAPPING
            )

        no_child_has_a_known_level = not any(has_node_a_known_level(x) for x in splits)
        # No child has a known level this means we only have sentences and no more semantic information
        if no_child_has_a_known_level:
            return DocumentNode(
                highest_level=highest_level,
                token_len=token_len,
                text=text,
                metadata=metadata,
                children=[
                    DocumentNode(
                        highest_level=highest_level,
                        token_len=_get_token_len(sent),
                        text=sent,
                        metadata=metadata,
                        children=[],
                    )
                    for sent in self.text_sentences(text)
                ],
            )

        # Further split the child nodes until we reach the token limit for each
        children: list[DocumentNode] = []
        for s in splits:
            if not hasattr(s, "children"):
                continue
            md_child = self._render_doc(s)
            highest_level_child, _, _ = self._find_highest_level(s.children)
            children.append(
                DocumentNode(
                    highest_level=highest_level_child,
                    text=md_child,
                    token_len=_get_token_len(md_child),
                    children=[
                        self._markdown_hierarchy_parser(
                            md_child, metadata, max_depth - 1
                        )
                    ],
                    metadata=metadata,
                )
            )
        return DocumentNode(
            highest_level=highest_level,
            token_len=token_len,
            text=text,
            children=children,
            metadata=metadata,
        )

    # unused
    def _split_by_sentence(self, text: str) -> list[str]:
        """sometimes _split_children does not find children leafs with are smaller then TOKEN_LIMIT.
        Thus we need to split by sentence"""
        token_limit = self.token_limit
        token_buffer = self.token_limit_buffer
        lenth = _get_token_len(text)
        needed_splits = lenth // token_limit
        if not needed_splits:
            return [text]
        sentences = [
            (_get_token_len(sent), f"{sent}. ")
            for sent in re.split(r"\.(?=\s|\\n)", text)
            if sent.strip()
        ]
        chunks: list[str] = []
        chunk = ""
        chunk_len = 0
        for size, sent in sentences:
            if size > token_limit + token_buffer:  # single big sentence
                if chunk:  # then add last
                    chunks.append(chunk)
                    chunk = ""
                    chunk_len = 0
                chunks.append(_cut_to_tokenlen(sent, token_limit))  # cut this
                # Last piece of sentence is discarded
                continue
            if chunk_len + size > token_limit + token_buffer:  # with next to big
                chunks.append(chunk)
                chunk = ""
                chunk_len = 0
            chunk += sent
            chunk_len += size
            if (
                token_limit + token_buffer >= chunk_len >= token_limit - token_buffer
            ):  # together they fit
                chunks.append(chunk)

                chunk = ""
                chunk_len = 0
        if chunk:
            chunks.append(_cut_to_tokenlen(chunk, token_limit))
        # chunks = [
        #    (chunk.replace("\n").strip() if not "#" else chunk.strip()) for chunk in chunks
        # ] This was broken
        chunks = [
            (
                _cut_to_tokenlen(chunk, token_limit)
                if not self._is_within_targetlen_w_buffer(chunk)
                or self._is_short(chunk)
                else chunk
            )
            for chunk in chunks
        ]
        for chunk in chunks:
            assert self._is_within_targetlen_w_buffer(chunk) or self._is_short(chunk)
        return chunks

    # pylint: disable-next=too-many-positional-arguments
    def _handle_parsing_of_children(
        self,
        doc: DocumentNode,
        child: DocumentNode,
        text_w_prev_child: str,
        remaining_snipped: str,
        recursive_depth: int,
    ) -> tuple[str, list[MarkdownDataContract]]:
        """Handle parsing of children in parse in _parse_hierarchical
        Outsourcing logic for the sole purpose of R0912: Too many branches (15/12) (too-many-branches)

        Args:
            doc (DocumentNode)
            child (DocumentNode)
            text_w_prev_child (str)
            remaining_snipped (str)
            recursive_depth (int): internal recursion-depth tracker

        Returns:
            tuple[str, list[DocumentNode]]: remaining_snipped, MdContracts
        """
        return_doc = []
        if self._is_short(text_w_prev_child):
            remaining_snipped = text_w_prev_child
        elif self._is_within_targetlen_w_buffer(text_w_prev_child):
            child["text"] = text_w_prev_child
            return_doc += [
                MarkdownDataContract(
                    md=_cut_to_tokenlen(child["text"], self.token_limit),
                    url=child["metadata"]["url"],
                    keywords=child["metadata"]["keywords"],
                )
            ]
            remaining_snipped = ""
        else:
            if not _is_standalone_a_heading(remaining_snipped):
                if _get_token_len(remaining_snipped) >= self.token_limit_min:
                    return_doc.append(
                        MarkdownDataContract(
                            md=remaining_snipped,
                            keywords=doc["metadata"]["keywords"],
                            url=doc["metadata"]["url"],
                        )
                    )
                remaining_snipped = ""
                if self._is_within_targetlen_w_buffer(child["text"]):
                    return_doc.append(self._md_data_from_dict_cut(child))
                else:
                    return_doc += self._parse_hierarchical(child, recursive_depth + 1)
            else:
                temp_docs = self._parse_hierarchical(child, recursive_depth + 1)
                return_doc += [
                    MarkdownDataContract(
                        md=remaining_snipped + "\n\n" + d.md,
                        keywords=d.keywords,
                        url=d.url,
                    )
                    for d in temp_docs
                ]
        return remaining_snipped, return_doc

    def _md_data_from_dict_cut(self, doc):
        return MarkdownDataContract(
            md=_cut_to_tokenlen(doc["text"], self.token_limit),
            url=doc["metadata"]["url"],
            keywords=doc["metadata"]["keywords"],
        )

    def _parse_hierarchical(
        self,
        doc: DocumentNode,
        recursive_depth: int = 1,
    ) -> list[MarkdownDataContract]:
        if _get_token_len(doc["text"]) <= self.token_limit_min:
            if recursive_depth == 1:
                return [self._md_data_from_dict_cut(doc)]
            log.warning("document to short", extra=doc)
            return []
        if self._is_within_targetlen_w_buffer(doc["text"]):
            return [self._md_data_from_dict_cut(doc)]
        if "children" not in doc.keys():
            log.warning(
                "no remaining children. still to big -> sentence split by dot",
                extra=doc,
            )
            return [self._md_data_from_dict_cut(doc)]
        if self._is_table(doc):
            log.warning(
                "found table, that should have been split, cutting off",
                extra=doc["metadata"],
            )
            return [self._md_data_from_dict_cut(doc)]
        if len(doc["children"]) == 0:
            log.warning(
                "no remaining children. still to big -> Cut by tokenlen",
                extra=doc,
            )
            return [self._md_data_from_dict_cut(doc)]
        if len(doc["children"]) == 1:
            return self._parse_hierarchical(doc["children"][0])

        children = doc["children"]
        return_doc: list[MarkdownDataContract] = []

        # If we don't have a proper hierarchy, handle the elements before top item separately
        if not _is_all_children_same_level(children):
            idx_highest = _get_highest_index_of_children(children)
            highest_child_is_heading = _get_highest_level_of_children(children) <= 6
            first_child_is_highest = idx_highest == 0
            if not first_child_is_highest and highest_child_is_heading:
                text_until_highest_child = "\n".join(
                    [c["text"] for c in children[:idx_highest]]
                )
                token_len = _get_token_len(text_until_highest_child)
                max_level = max(c["highest_level"] for c in children[:idx_highest])
                new_doc = DocumentNode(
                    highest_level=max_level,
                    token_len=token_len,
                    text=text_until_highest_child,
                    children=children[:idx_highest],
                    metadata=doc["metadata"],
                )
                return_doc += self._parse_hierarchical(new_doc)
                children = children[idx_highest:]

        remaining_snipped = ""
        for child in children:
            child["metadata"] = doc["metadata"]  # inherit metadata downwards
            text_w_prev_child = "\n".join([remaining_snipped, child["text"]]).strip()
            remaining_snipped, returned_docs = self._handle_parsing_of_children(
                doc, child, text_w_prev_child, remaining_snipped, recursive_depth
            )
            return_doc += returned_docs

        # add potential short remaining spillovers
        if _get_token_len(remaining_snipped) >= self.token_limit_min:
            return_doc += [
                MarkdownDataContract(
                    md=_cut_to_tokenlen(remaining_snipped, self.token_limit),
                    url=doc["metadata"]["url"],
                    keywords=doc["metadata"]["keywords"],
                )
            ]
        return return_doc

    def _adopt_splitted_list_to_use_highest_prev_header(
        self,
        docs: list[MarkdownDataContract],
    ) -> list[MarkdownDataContract]:
        """
        Function to improve the semantic meaning of the Markdown document by reattaching a parent heading

        Does not yet respect the token limit, however headings usually have little impact
        """
        highest_header_until_now = {i + 1: "" for i in range(6)}
        for doc in docs:
            text = doc.md.strip()
            md_doc = MisDocument(text)
            # Create Header
            highest_level, highest_type, highest_child = self._find_highest_level(
                md_doc.children
            )
            if str(highest_type) == str(block_token.Heading):
                # Discuss this
                # highest_header_until_now[highest_level] = _get_heading_text(highest_child)
                highest_header_until_now[highest_level] = self._render_doc(
                    highest_child
                ).lstrip(" #")

            ordered_headers = {
                level: text for level, text in highest_header_until_now.items() if text
            }
            docwide_highest_level = (
                min(ordered_headers.keys()) if ordered_headers else 10
            )
            new_header = " - ".join(
                [
                    text
                    for level, text in ordered_headers.items()
                    if level < highest_level
                ]
            )
            if new_header:
                new_header = "# " + new_header
            # Filter Doc
            document_is_just_single_header = (
                text.strip().startswith("#") and "\n" not in text.strip()
            )
            if document_is_just_single_header:
                continue
            # TODOo check token limit or limit header lenght
            # The higher heading the lower its level
            doc_has_lower_heading = docwide_highest_level < highest_level
            new_doc = (
                text.strip()
                if not doc_has_lower_heading
                else (new_header + "\n\n" + doc.md).strip()
            )
            doc.md = new_doc

        return docs

    def split_markdown_document(
        self, doc: MarkdownDataContract
    ) -> list[MarkdownDataContract]:
        """Split a Markdown Document into Snippets"""
        metadata = MetaDataDict(url=doc.url, keywords=doc.keywords)
        doc_hierarchy: DocumentNode = self._markdown_hierarchy_parser(doc.md, metadata)
        doc_snippets: list[MarkdownDataContract] = self._parse_hierarchical(
            doc_hierarchy
        )
        improved_snippets: list[MarkdownDataContract] = (
            self._adopt_splitted_list_to_use_highest_prev_header(doc_snippets)
        )
        formatted_snippets: list[MarkdownDataContract] = _format_markdown_docs(
            improved_snippets
        )
        return formatted_snippets



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

from wurzel.adapters.dvc_adapter import Backend
from wurzel.step.typed_step import TypedStep


def main(step: TypedStep, data_dir: Path, backend: Backend) -> str:
    """
    Generates the yaml for the given backend
    """

    adapter: Backend = backend(data_dir)
    return adapter.generate_yaml(step)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from wurzel.utils import HAS_QDRANT as _HAS_QDRANT

from ..step import TypedStep
from .embedding import EmbeddingStep
from .manual_markdown import ManualMarkdownStep

__all__ = ["TypedStep", "ManualMarkdownStep", "EmbeddingStep"]
if _HAS_QDRANT:
    from .qdrant import QdrantConnectorMultiVectorStep, QdrantConnectorStep

    __all__ += [
        "QdrantConnectorMultiVectorStep",
        "QdrantConnectorStep",
    ]
